/*	nkfs for Linux
 *	
 *	Copyright (c) 1995, 1999  Caldera Systems, Inc.
 *	
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *	Boston, MA  02111-1307, USA.
 *
 * 	Please submit patches to nkfs@calderasystems.com
 *
 */

#define KERNEL_22

#define __NO_VERSION__
#include <linux/modversions.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/fcntl.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/signal.h>
#include <nkfs.h>

static struct file  	*Gfp;

static nkfs_driver_t	*nkfs_debug_list[10];
static int		nkfs_debug_cnt = 0;
#ifdef DEBUG
static unsigned long	nkfs_debug_print = 1;
#else
static unsigned long	nkfs_debug_print = 0;
#endif

int
nkfs_register_notifier(nkfs_driver_t *dp, void (*fp)(void *), void *arg)
{
	if ((dp == NULL) || (fp == NULL))
		return -EINVAL;

	dp->callback = fp;
	dp->arg = arg;
	return 0;
}

int
nkfs_unregister_notifier(nkfs_driver_t *dp)
{
	if (dp == NULL)
		return -EINVAL;
	dp->callback = NULL;
	return 0;
}

nkfs_rqst_t *
make_nkfs_request(long size)
{
	nkfs_rqst_t	*rp = NULL;

	rp = (nkfs_rqst_t *) kmalloc(sizeof(nkfs_rqst_t)+size, GFP_ATOMIC);
	return rp;
}

nkfs_rply_t *
make_nkfs_reply(long size)
{
	nkfs_rply_t	*rp;

#ifdef DEBUG
	printk("NKFS: reply() size=%08X \n",size);
#endif
	rp = (nkfs_rply_t *) kmalloc(sizeof(nkfs_rply_t)+size, GFP_ATOMIC);
	return rp;
}

void
nkfs_send(nkfs_driver_t *driver, nkfs_rqst_t *rqst, unsigned long size, 
		unsigned long type)
{
	nkfs_rqstq_t	*outp;
	nkfs_rqstq_t	*rqp;
#ifdef DEBUG
	printk("NKFS: send() size=%08X \n",size);
#endif
	

	if (rqst == NULL) return;
	if ((driver == NULL) || (driver->signature != NKFS_DRIVER_SIG)) {
		kfree(rqst);
		return;
	}

	/* Fill in request */
	rqst->hdr.msglen = size + (int)(rqst->data - (unsigned char *)rqst);
	rqst->hdr.msgtype = type;
	rqst->hdr.uid = current->euid;
	rqst->hdr.gid = current->egid;
	rqst->hdr.seqnum = (driver->seqnum)++;

	outp = (nkfs_rqstq_t *)kmalloc(sizeof(*outp), GFP_ATOMIC);
	if (outp == NULL) {
		kfree(rqst);
		return;
	}

	outp->rqst = rqst;
	outp->next = NULL;
	outp->free = 1;

	if (driver->outgoing == NULL) {
		driver->outgoing = outp;
		outp->prev = NULL;
	} else {
		for (rqp = driver->outgoing;  rqp->next; rqp = rqp->next);
		rqp->next = outp;
		outp->prev = rqp;
	}

	/* Wake up reader */
	if (waitqueue_active(&(driver->wait)))
		wake_up_interruptible(&(driver->wait));

	return;
}

#define BITNUM(bit)	(1<<((bit)-1))
nkfs_rply_t *
nkfs_rpc(nkfs_driver_t *driver, nkfs_rqst_t *rqst, unsigned long size, 
		unsigned long type)
{
	nkfs_rplyq_t	pending;
	nkfs_rqstq_t	*outp;
	nkfs_rqstq_t	*rqp;
	nkfs_rplyq_t	*rpp;
	sigset_t	saved_mask;
	unsigned long 	flags, sigpipe;
#ifdef DEBUG
	printk("NKFS: rpc() size =%d \n",size);
#endif

	if (rqst == NULL) {
		if (nkfs_debug_print) 
			printk("NKFS: rpc, null request: %x\n", (int) rqst);
		return NULL;
	}
	if ((driver == NULL) || (driver->signature != NKFS_DRIVER_SIG)) {
		kfree(rqst);
		if (nkfs_debug_print) 
			printk("NKFS: rpc, driver=NULL or dsig bad: %x\n", (int) rqst);
		return NULL;
	}

	/* Fill in request */
	rqst->hdr.msglen = size + (int)(rqst->data - (unsigned char *)rqst);
	rqst->hdr.msgtype = type;
	rqst->hdr.uid = current->euid;
	rqst->hdr.gid = current->egid;

	outp = (nkfs_rqstq_t *)kmalloc(sizeof(*outp), GFP_ATOMIC);
	if (outp == NULL) {
		kfree(rqst);
		if (nkfs_debug_print) 
			printk("NKFS: rpc, outp=NULL : %x\n", (int) rqst);
		return NULL;
	}

	outp->rqst = rqst;
	outp->next = NULL;
	outp->free = 1;

	/* Fill in pending reply */
	pending.seqnum = rqst->hdr.seqnum = (driver->seqnum)++;
	pending.next = NULL;
	init_waitqueue(&(pending.wait));
	pending.rply = NULL;

	/* Queue request and pending reply */
	if (driver->incoming == NULL) {
		driver->incoming = &pending;
		pending.prev = NULL;
	} else {
		for (rpp = driver->incoming; rpp->next; rpp = rpp->next);
		rpp->next = &pending;
		pending.prev = rpp;
	}

	if (driver->outgoing == NULL) {
		driver->outgoing = outp;
		outp->prev = NULL;
	} else {
		for (rqp = driver->outgoing;  rqp->next; rqp = rqp->next);
		rqp->next = outp;
		outp->prev = rqp;
	}

	/* Mask signals */
// 
// This is a whole new chunk of code for the 2.2 Kernel to safely handle
// critical sections on SMP machines
// 
#ifdef DEBUG
	printk("NKFS: rpc, about to enter critical section \n"); 
#endif
	spin_lock_irqsave(&current->sigmask_lock, flags);
	sigpipe = sigismember(&current->signal, SIGPIPE);
	saved_mask = current->blocked;
	siginitsetinv(&current->blocked, sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGQUIT));
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, flags);
#ifdef DEBUG
	printk("NKFS: rpc, IN critical section \n"); 
#endif
	
	/* Wake up reader */
	if (waitqueue_active(&(driver->wait)))
	{
		wake_up_interruptible(&(driver->wait));
	}

#ifdef DEBUG
	printk("NKFS: rpc, blocking for response \n"); 
#endif
	/* Wait for reply */
	interruptible_sleep_on(&(pending.wait));
	

#ifdef DEBUG
	printk("NKFS: rpc, restoring signals\n"); 
#endif
	spin_lock_irqsave(&current->sigmask_lock, flags);
	if (!sigpipe)
		sigdelset(&current->signal, SIGPIPE);
	current->blocked = saved_mask;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->signask_lock, flags);

	if (pending.prev) {
		pending.prev->next = pending.next;
	} else {
		driver->incoming = pending.next;
	}
	if (pending.next) {
		pending.next->prev = pending.prev;
	}

	if ((driver->state == NKFS_DRIVER_DRAINING) &&
			(driver->incoming == NULL) &&
			waitqueue_active(&(driver->wait)))
		wake_up_interruptible(&(driver->wait));

	if (nkfs_debug_print) 
		printk("NKFS: rpc, return: %x\n", (int) rqst);
	return pending.rply;
}


static nkfs_driver_t *
make_nkfs_driver(void)
{
	nkfs_driver_t	*dp;

	dp = (nkfs_driver_t *) kmalloc(sizeof(nkfs_driver_t), GFP_ATOMIC);
	if (dp) {
		dp->signature = NKFS_DRIVER_SIG;
		dp->state = NKFS_DRIVER_NORMAL;
		dp->incoming = NULL;
		dp->outgoing = NULL;
		dp->callback = NULL;
		dp->arg = NULL;
		init_waitqueue(&(dp->wait));
		dp->seqnum = 0L;
		if (++nkfs_debug_cnt > 9) {
			printk("NKFS: driver, Unable to save debug pointer.\n");
		} else {
			nkfs_debug_list[nkfs_debug_cnt] = dp;
		}
	}

	return dp;
}

static void
free_nkfs_driver(nkfs_driver_t *dp)
{
	nkfs_rplyq_t	*rpp;
	nkfs_rqstq_t	*rqp;
	int		i;

	if (dp->callback)
		(dp->callback)(dp->arg);

	for (rpp = dp->incoming; rpp; rpp = rpp->next) {
		if (waitqueue_active(&(rpp->wait)))
			wake_up_interruptible(&(rpp->wait));
	}

	for (rqp = dp->outgoing; rqp; ) {
		nkfs_rqstq_t	*tmp = rqp;
		rqp = rqp->next;

		kfree(tmp->rqst);
		kfree(tmp);
	}

	dp->state = NKFS_DRIVER_DRAINING;
	while (dp->incoming)
		interruptible_sleep_on(&(dp->wait));

	for (i = 0; i < 10; i++) {
		if (nkfs_debug_list[i] == dp) {
			nkfs_debug_list[i] = NULL;
			nkfs_debug_list[i] = nkfs_debug_list[nkfs_debug_cnt--];
			break;
		}
	}
	kfree(dp);
}

// 
// function prototypes for nkfs driver calls
// 
static ssize_t nkfsd_read(struct file *, char *, size_t, loff_t *);
static ssize_t nkfsd_write(struct file *, const char *, size_t, loff_t *);

static unsigned int nkfsd_poll(struct file *, poll_table *);

static int nkfsd_ioctl(struct inode *, struct file *, unsigned int, unsigned long);

static int nkfsd_open(struct inode *, struct file *);

static int nkfsd_close( struct inode *, struct file *);


struct file_operations nkfs_driver_ops = {
	NULL,		/* lseek		*/
	nkfsd_read,	/* read 		*/
	nkfsd_write,	/* write 		*/
	NULL,		/* readdir		*/
	nkfsd_poll,	/* poll 	 	*/
	nkfsd_ioctl,	/* ioctl		*/
	NULL,		/* mmap			*/
	nkfsd_open,	/* open			*/
	NULL,		/* flush		*/
	nkfsd_close,	/* release		*/
	NULL,		/* fsync		*/
	NULL,		/* fasync		*/
	NULL,		/* chekc_media_change	*/
	NULL,		/* revalidate		*/
	NULL		/* lock */		
};

static ssize_t 
nkfsd_read(struct file *fp, char *buf, size_t len, loff_t *off)
{
	nkfs_driver_t	*dp = (nkfs_driver_t *)(fp->private_data);
	nkfs_rqstq_t 	*rqp;
	nkfs_rqst_t 	*rqst;
	int		copied;
	int		fullsize;
	int		err;

	err = verify_area(VERIFY_WRITE, buf, len);
	if (err) return err;

	if ((dp == NULL) || (dp->signature != NKFS_DRIVER_SIG)) {
		printk("NKFS: read, Invalid driver\n");
		return -EINVAL;
	}

	/* Find first available */
	while (dp->outgoing == NULL) {
		if (fp->f_flags & O_NONBLOCK) {
			if (nkfs_debug_print) 
				printk("NKFS: read, Would block.\n");
			return -EWOULDBLOCK;
		}

		if ( signal_pending(current)) {
			return -EINTR;
		}

		interruptible_sleep_on(&(dp->wait));
	}

	rqp = dp->outgoing;
	rqst = rqp->rqst;
	rqp->rqst = NULL;
	if (rqp->next)
		rqp->next->prev = NULL;
	dp->outgoing = rqp->next;

	if (rqp->free) kfree(rqp);
	if (rqst == NULL) return 0;

	fullsize = rqst->hdr.msglen;
	copied = (fullsize > len) ? len : fullsize;
	copy_to_user(buf, rqst, copied);

	kfree(rqst);

	return fullsize;
}

ssize_t 
nkfsd_write(struct file *fp, const char *buf, size_t len, loff_t *offset)
{
	nkfs_driver_t	*dp = (nkfs_driver_t *)(fp->private_data);
	nkfs_rplyq_t	*rpp;
	nkfs_rply_t	*rply;
	int		err;

#ifdef DEBUG
	printk("NKFS: write\n");
#endif
	err = verify_area(VERIFY_READ, buf, len);
	if (err) return err;

	if ((dp == NULL) || (dp->signature != NKFS_DRIVER_SIG))
		return -EINVAL;

	rply = (nkfs_rply_t *)kmalloc(len, GFP_ATOMIC);
	if (rply == NULL) return -ENOMEM;
	copy_from_user(rply, buf, len);

	for (rpp = dp->incoming; 
			rpp && (rply->hdr.seqnum != rpp->seqnum);
			rpp = rpp->next);

	if (rpp == NULL) {
		kfree(rply);
		return len;
	}

	rpp->rply = rply;
	if (waitqueue_active(&(rpp->wait)))
		wake_up_interruptible(&(rpp->wait));

	return len;
}

// Poll is implemented instead of select() in the 2.2 Linux Kernel

static unsigned int 
nkfsd_poll(struct file *fp, poll_table *pt)
{
int flag = SEL_IN;
struct poll_table_entry;
nkfs_driver_t	*dp = (nkfs_driver_t *)(fp->private_data);

	if ((dp == NULL) || (dp->signature != NKFS_DRIVER_SIG))
		return -EINVAL;

	if (flag != SEL_IN) return 1;
	if (dp->outgoing) return 1;
	
	poll_wait(fp, &(dp->wait),pt);

	return 0;
}

static void
nkfs_print_incoming(nkfs_driver_t *dp)
{
	nkfs_rplyq_t	*rpp;

	printk("\tIncoming:");
	for (rpp = dp->incoming; rpp; rpp = rpp->next) {
		printk(" %lx", rpp->seqnum);
	}
	printk("\n");
} 

static void
nkfs_print_outgoing(nkfs_driver_t *dp)
{
	nkfs_rqstq_t 	*rqp;

	printk("\tOutgoing:");
	for (rqp = dp->outgoing; rqp; rqp = rqp->next) {
		if (rqp->rqst)
			printk(" %lx", rqp->rqst->hdr.seqnum);
		else 
			printk(" NULL");
	}
	printk("\n");
}

static int 
nkfsd_ioctl(struct inode *ip, struct file *fp, unsigned int cmd, unsigned long arg)
{
	nkfs_driver_t	*dp;

#ifdef DEBUG
	printk("NKFS: ioctl \n");
#endif

	switch (cmd) {
	case 0:
		printk("NKFS: ioctl, Dumping entry %lx\n", arg);
		dp = nkfs_debug_list[arg];
		if (dp) {
			nkfs_print_incoming(dp);
			nkfs_print_outgoing(dp);
		}
		printk("NKFS: ioctl, Completed dumping entry %lx\n", arg);
		break;
	
	case 1:
		nkfs_debug_print = arg; 
		break;
	}
	return 0;
}

static int 
nkfsd_open(struct inode *ip, struct file *fp)
{
	nkfs_driver_t	*dp;
#ifdef DEBUG
	printk("NKFS: nkfsd_open: ip=%08X, fp=%08X\n", ip, fp);
	printk("NKFS: nkfsd_open: ip->sb=%08X\n", ip->i_sb);
#endif

	MOD_INC_USE_COUNT;
	dp = make_nkfs_driver();
	Gfp = fp;
	if (dp == NULL) {
	       	return -ENOMEM;
	}
	fp->private_data = dp;
	return 0;
}

static int 
nkfsd_close(struct inode *ip, struct file *fp)
{
	nkfs_driver_t	*dp;
#ifdef DEBUG
	printk("NKFS: close \n");
#endif

	if ((fp == NULL) || (fp->private_data == NULL)) {
		MOD_DEC_USE_COUNT;
		return(0);
	}

	dp = (nkfs_driver_t *)(fp->private_data);
	if (dp->signature == NKFS_DRIVER_SIG)
		free_nkfs_driver(dp);
	fp->private_data = NULL;
	MOD_DEC_USE_COUNT;

	return(0);
}

static int	major = -1;

int 
init_nkfsd(void)
{
	major = register_chrdev(0, NKFS_DRIVER_NAME, &nkfs_driver_ops);
	if (major < 0)
		return major;
	return 0;
}

void
cleanup_nkfsd(void)
{
	if (major > 0)
		(void) unregister_chrdev(major, NKFS_DRIVER_NAME);
	major = -1;
}

