/*	nkfs for Linux
 *	
 *	Copyright (c) 1995, 1999  Caldera Systems, Inc.
 *	
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *	Boston, MA  02111-1307, USA.
 *
 * 	Please submit patches to nkfs@calderasystems.com
 *
 */

#define __NO_VERSION__
#include <linux/modversions.h>

#include <linux/module.h>
#include <linux/version.h>
#include <linux/stat.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/locks.h>
#include <asm/uaccess.h>
#include <nkfs_mount.h>
#include <linux/fs.h>
#include <linux/signal.h>
#include <nkfs.h>

extern struct file *Gfp;

extern struct inode_operations nkfs_inode_ops;

static int nkfs_notify_change(struct dentry *, struct iattr *);
static void nkfs_read_inode(struct inode *);
static void nkfs_put_inode(struct inode *);
static void nkfs_delete_inode(struct inode *);
static void nkfs_put_super(struct super_block *);
static int nkfs_statfs(struct super_block *, struct statfs *, int);

struct super_operations	nkfs_super_ops = {
	nkfs_read_inode,	/* read_inode		*/
	NULL,			/* write_inode		*/
	nkfs_put_inode,		/* put_inode		*/
	nkfs_delete_inode,	/* delete_inode		*/
	nkfs_notify_change,	/* notify_change	*/
	nkfs_put_super,		/* put_super		*/
	NULL,			/* write_super		*/
	nkfs_statfs,		/* statfs		*/
	NULL,			/* remount_fs		*/
	NULL,			/* clear_inode		*/
	NULL			/* umount_begin		*/
};

extern int
nkfs_register_notifier(nkfs_driver_t *, void (*fp)(void *), void *);

extern int nkfs_unregister_notifier(nkfs_driver_t *);

extern nkfs_rqst_t *make_nkfs_request(long);
extern nkfs_rply_t *nkfs_rpc(nkfs_driver_t *, nkfs_rqst_t *, unsigned long, unsigned long);

static ssize_t nkfs_read( struct file *, char *, size_t, loff_t *lf);
static ssize_t nkfs_write(struct file *, const char *, size_t, loff_t *lf);
static int nkfs_readdir(struct file *, void *, filldir_t);
static int nkfs_mmap(struct file *, struct vm_area_struct *);
static int nkfs_open(struct inode *, struct file *);
static int nkfs_release_file(struct inode *, struct file *);

struct file_operations nkfs_file_ops = {
	NULL,			// lseek
	nkfs_read,		// read
	nkfs_write,		// write
	nkfs_readdir,		// readdir
	NULL,			// poll
	NULL,			// ioctl
	nkfs_mmap,		// mmap
	nkfs_open,		// open
	NULL,			// flush
	nkfs_release_file, // release
	NULL,			// fsync	
	NULL,			// fasync
	NULL,			// check_media_change
	NULL,			// revalidate
	NULL			// lock
};

static int nkfs_create(struct inode *, struct dentry *, int);
static int nkfs_lookup(struct inode *, struct dentry *);
static int nkfs_mkdir(struct inode *, struct dentry *, int);
static int nkfs_rmdir(struct inode *, struct dentry *);
static int nkfs_unlink(struct inode *, struct dentry *);
static int nkfs_rename(struct inode *, struct dentry *,
	       	struct inode *, struct dentry *);
static int nkfs_readlink(struct dentry *, char *, int);
static struct dentry * nkfs_follow_link(struct dentry *, 
					struct dentry *, 
					unsigned int );
static void nkfs_truncate(struct inode *);
static int nkfs_permission(struct inode *, int);

struct inode_operations	nkfs_inode_ops = {
	&nkfs_file_ops,  	// Default file ops
	nkfs_create,     	// Create
	nkfs_lookup,     	// lookup
	NULL,		 	// link
	nkfs_unlink,     	// unlink 
	NULL,            	// symlink
	nkfs_mkdir,      	// mkdir
	nkfs_rmdir,      	// rmdir
	NULL,            	// mknod
	nkfs_rename,		// rename
	nkfs_readlink,		// readlink
	nkfs_follow_link,	// follow_link
	NULL,			// readpage
	NULL,			// writepage
	NULL,			// bmap
	nkfs_truncate,		// truncate
	nkfs_permission,	// permission
	NULL,			// smap
	NULL,			// updatepage
	NULL			// revalidate
};

/*
 * Dentry operations routines
 */
static int nkfs_lookup_validate(struct dentry *);
static int nkfs_hash_dentry(struct dentry *, struct qstr *);
static int nkfs_compare_dentry (struct dentry *, struct qstr *, struct qstr *);
static void nkfs_delete_dentry(struct dentry *);
static void nkfs_release_dentry(struct dentry *);
static void nkfs_iput_dentry(struct dentry *, struct inode *);

struct dentry_operations nkfs_dentry_operations =
{
	nkfs_lookup_validate,		// d_validate(struct dentry *)
	NULL, //nkfs_hash_dentry,	// d_hash
	NULL, //nkfs_compare_dentry,   	// d_compare
	NULL, //nkfs_delete_dentry,	// d_delete(struct dentry *)
	NULL, //nkfs_release_dentry,	// d_release(struct dentry *)
	NULL, //nkfs_iput_dentry        // d_iput(struct dentry *, struct inode *)
};

static int nkfs_lookup_validate(struct dentry *dentry)
{
#ifdef DEBUG
	printk("NKFS: dentry_operations: validate\n");
#endif
	if (!dentry->d_inode) {
		return 0;
	}
	if (dentry->d_inode->u.generic_ip == 0) {
		return 0;
	}
	if (dentry->d_time == 0) {
		return 0;
	}
	return 1;
}

#define tolower(c) (((c) >= 'A' && (c) <= 'Z') ? (c)-'A'+'a' : (c))

/*
 * Note: leave the hash unchanged if the directory
 * is case-sensitive.
 */
static int 
nkfs_hash_dentry(struct dentry *dentry, struct qstr *this)
{
	unsigned long hash;
	int i;

	hash = init_name_hash();
	for (i=0; i<this->len ; i++)
		hash = partial_name_hash(tolower(this->name[i]),hash);
	this->hash = end_name_hash(hash);
	return 0;
}

static int
nkfs_compare_dentry(struct dentry *dentry, struct qstr *a, struct qstr *b)
{
	int i;

	if (a->len != b->len) return 1;

	for (i=0; i<a->len; i++)
		if (tolower(a->name[i]) != tolower(b->name[i]))
			return 1;

	return 0;
}

/*
 * This is the callback from dput() when d_count is going to 0.
 * We use this to unhash dentries with bad inodes and close files.
 */
static void
nkfs_delete_dentry(struct dentry * dentry)
{
#ifdef DEBUG
	printk("NKFS: dentry_operations: delete, line %d\n", __LINE__);
#endif
	
	if (dentry->d_count == 1) {
		struct inode *inode = dentry->d_inode;
		if (inode) {
			dentry->d_inode = NULL;
			list_del(&dentry->d_alias);
			INIT_LIST_HEAD(&dentry->d_alias);
			if (dentry->d_op && dentry->d_op->d_iput) {
				dentry->d_op->d_iput(dentry, inode);
			} else {
				iput(inode);
			}
		}
		return;
	}
	d_drop(dentry);
}

static void nkfs_release_dentry(struct dentry *dentry)
{
#ifdef DEBUG
	printk("NKFS: dentry_operations: release_dentry, line %d\n", __LINE__);
#endif
}

static void nkfs_iput_dentry(struct dentry *dentry, struct inode *inode)
{
#ifdef DEBUG
	printk("NKFS: dentry_operations: iput_dentry, line %d\n", __LINE__);
#endif
}


void
nkfs_refresh_inode(struct inode *inode, nkfs_node_info_t *nodeinfo)
{
	int	blksize;

#ifdef DEBUG
	printk("NKFS: refresh inode  \n"); 
#endif

	if ((inode == NULL) || (nodeinfo == NULL))
	{
		printk("NKFS: refresh_node, return NULL inode! \n");
		return;
	}

	inode->i_mode = nodeinfo->mode;
	inode->i_uid = nodeinfo->uid;
	inode->i_gid = nodeinfo->gid;
	inode->i_size = nodeinfo->size;
	inode->i_atime = nodeinfo->atime;
	inode->i_mtime = nodeinfo->mtime;
	inode->i_ctime = nodeinfo->ctime;
	inode->i_nlink = nodeinfo->nlink;
	blksize = inode->i_sb->s_blocksize;
	inode->i_blksize = blksize;
	if (blksize)
		inode->i_blocks = (inode->i_size + blksize - 1) / blksize;
}

static int
nkfs_unlock_node(nkfs_super_block_t *nkfssb, unsigned long nodeid)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_unlock_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	int	reqlen = sizeof(nkfs_unlock_rqst_t);
	int	result;

#ifdef DEBUG
	printk("NKFS: unlock \n"); 
#endif

	if (((void *)nkfssb == NULL) || (nkfssb->driver == NULL)) 
	{
		printk("NKFS: unlock, sb->driver = NULL \n");
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nodeid;

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_UNLOCKNODE);

	if (rply)
		result = rply->hdr.result;
	else
	{
		printk("NKFS: unlock, rply = NULL \n");
		result = -EINTR;
	}

	if (rply) kfree(rply);

	return result;
}

struct inode *
nkfs_make_inode(struct super_block *sb, nkfs_node_info_t *nodeinfo)
{
	nkfs_inode_t	*nkfsi;
	struct inode	*inode;

#ifdef DEBUG
	printk("NKFS: make, sb=%08X, nodeinfo=%08X \n", sb, nodeinfo); 
#endif
	if (nodeinfo == NULL)
	{
		printk("NKFS: make, NodeInfo is EMPTY! \n");
		return(NULL);
	}

	inode = iget(sb, nodeinfo->ino);
	
	if (inode == NULL) {
		printk("NKFS: make, Unable to allocate generic inode.\n");
		return NULL;
	} 
	
	if (inode->i_count == 1)
	{
		nkfsi = (nkfs_inode_t *)kmalloc(sizeof(nkfs_inode_t), 
			GFP_ATOMIC);
		if (nkfsi == NULL) {
			printk("NKFS: make, Unable to allocate inode.\n");
			iput(inode);
			return NULL;
		}
	
		inode->u.generic_ip = nkfsi;
		inode->i_nlink = 1;
		inode->i_blksize = 512;
		inode->i_rdev = 0;
		inode->i_blocks = 0;
		nkfsi->inode = inode;
		nkfsi->nodeid = nodeinfo->nodeid;
		nkfsi->nkfssb = (nkfs_super_block_t *)(sb->u.generic_sbp);
	} else if (inode->i_count < 1) {
		printk("NKFS: make, returning put inode.\n");
	}
		
	nkfs_refresh_inode(inode, nodeinfo);

	return inode;
}
	
static int nkfs_create(struct inode *ip, struct dentry *de, int mode)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_create_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_create_rply_t	msg;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_create_rqst_t);
	int	result;
	int len = de->d_name.len;	
	struct inode *newip = NULL;	

#ifdef DEBUG
	printk("NKFS: create\n"); 
#endif

	if (ip == NULL) return -ENOENT;
	if (!S_ISDIR(ip->i_mode) || (ip->u.generic_ip == NULL)) {
		printk("NKFS: create, No generic_ip \n");
		iput(ip);
		return -ENOENT;
	}

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		printk("NKFS: create, driver or sb = NULL! \n");
		iput(ip);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	rqst->msg.mode = mode;
	memcpy(rqst->msg.name, de->d_name.name,len );
	rqst->msg.name[len] = '\0';

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_CREATE);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (result == 0) {
		newip = nkfs_make_inode(ip->i_sb, &(rply->msg.nodeinfo));
		d_instantiate(de, newip);
	}

	if (rply) kfree(rply);

	if (ip->i_count < 1) {
		printk("NKFS: create, Reputting inode.\n");
	}
	
//	iput(ip);	// results in "still has aliases" error
	return result;
}

static int nkfs_lookup(struct inode *dir, struct dentry *de)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_lookup_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_lookup_rply_t	msg;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_lookup_rqst_t);
	int	result;
	
	int	len = de->d_name.len;
	struct inode **newip = &(de->d_inode);

#ifdef DEBUG
	printk("NKFS: lookup \n"); 
	printk("NKFS lookup: dir=%08X\n", dir); 
	printk("NKFS lookup: de=%08X\n", de); 
	printk("NKFS lookup: de->name=%s\n", de->d_name.name); 
#endif

	*newip = NULL;
	if (dir == NULL) return -ENOENT;
	if (!S_ISDIR(dir->i_mode) || (dir->u.generic_ip == NULL)) {
		printk("NKFS: lookup, No generic ip.\n");
		iput(dir);
		return -ENOENT;
	}

	nkfsi = (nkfs_inode_t *)(dir->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		printk("NKFS: lookup, No connection.\n");
		iput(dir);
		return -EIO;
	}

	if ((de->d_name.name[0] == '.') && (len == 1)) {
		if (dir->i_count < 1) {
			printk("NKFS: lookup, returning self.\n");
		}
		*newip = dir;
		return 0;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	memcpy(rqst->msg.name, de->d_name.name, len);
	rqst->msg.name[len] = '\0';

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_LOOKUP);

	if (rply)
		result = rply->hdr.result;
	else
	{
		printk("NKFS: lookup, rply = NULL \n");
		result = -EINTR;
	}

	if (result == 0) {
		*newip = nkfs_make_inode(dir->i_sb, &(rply->msg.nodeinfo));
		d_add(de, *newip);
	} else {
		d_add(de, NULL);		// negative dentry
	}
	de->d_op = &nkfs_dentry_operations;
	(void) nkfs_unlock_node(nkfssb, rply->msg.nodeinfo.nodeid);

	if (rply) kfree(rply);
	
//	iput(dir);
	return 0;
}

static int nkfs_unlink(struct inode *ip, struct dentry *de)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_unlink_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_unlink_rqst_t);
	int	result;
	
	int 	len = de->d_name.len;
	
#ifdef DEBUG
	printk("NKFS: unlink \n"); 
#endif

	if (ip == NULL) return -ENOENT;
	if (!S_ISDIR(ip->i_mode) || (ip->u.generic_ip == NULL)) {
		iput(ip);
		return -ENOENT;
	}

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		iput(ip);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	memcpy(rqst->msg.name, de->d_name.name, len);
	rqst->msg.name[len] = '\0';

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_UNLINK);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (rply) kfree(rply);

//	iput(ip); // results in "semaphore in use", "aliases" errors
	return result;
}

ino_t generate_ino(unsigned long n)
{
	static ino_t ino = 2; 
	
	if (ino + 2*n < ino) { 
		ino = 2; 
	}
	ino += n;
	return ino;
}

static int nkfs_mkdir(struct inode *ip, struct dentry *de, int mode)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_mkdir_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_mkdir_rqst_t);
	int	result;

	int 	len = de->d_name.len;
	struct  inode *inode;


#ifdef DEBUG
	printk("NKFS: mkdir \n"); 
#endif

	if (ip == NULL) return -ENOENT;
	if (!S_ISDIR(ip->i_mode) || (ip->u.generic_ip == NULL)) {
		iput(ip);
		return -ENOENT;
	}

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		iput(ip);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	rqst->msg.mode = mode;
	memcpy(rqst->msg.name, de->d_name.name, len);
	rqst->msg.name[len] = '\0';

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_MKDIR);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (rply) kfree(rply);

//	iput(ip);

	inode = iget(nkfsi->nkfssb->sb, generate_ino(1));
	if (inode != NULL) { 
		inode->i_mode |= 0040000;
		d_instantiate(de, inode); 
	}
	
	return result;
}

static int nkfs_rmdir(struct inode *ip, struct dentry *de)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_rmdir_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_rmdir_rqst_t);
	int	result;
	int 	len = de->d_name.len;

#ifdef DEBUG
	printk("NKFS: rmdir \n"); 
#endif

	if (ip == NULL) return -ENOENT;
	if (!S_ISDIR(ip->i_mode) || (ip->u.generic_ip == NULL)) {
		iput(ip);
		return -ENOENT;
	}

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		iput(ip);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	memcpy(rqst->msg.name, de->d_name.name, len);
	rqst->msg.name[len] = '\0';

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_RMDIR);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (rply) kfree(rply);

//	iput(ip); // causes semaphore, aliases error messages
	return result;
}

static int nkfs_rename(struct inode *odir, struct dentry *ode,
	       	struct inode *ndir, struct dentry *nde)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_rename_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*onkfsi;
	nkfs_inode_t		*nnkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_rename_rqst_t);
	int	result;

#ifdef DEBUG
	printk("NKFS: rename \n"); 
#endif

	if ((odir == NULL) || (ndir == NULL)) {
		if (odir) iput(odir);
		if (ndir) iput(ndir);
		return -ENOENT;
	}
	if (!S_ISDIR(odir->i_mode) || (odir->u.generic_ip == NULL)) {
		iput(odir);
		iput(ndir);
		return -ENOENT;
	}

	if (!S_ISDIR(ndir->i_mode) || (ndir->u.generic_ip == NULL)) {
		iput(odir);
		iput(ndir);
		return -ENOENT;
	}

	onkfsi = (nkfs_inode_t *)(odir->u.generic_ip);
	nnkfsi = (nkfs_inode_t *)(ndir->u.generic_ip);

	if (((nkfssb = onkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		iput(odir);
		iput(ndir);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.onodeid = onkfsi->nodeid;
	rqst->msg.nnodeid = nnkfsi->nodeid;
	memcpy(rqst->msg.oname, ode->d_name.name, ode->d_name.len);
	rqst->msg.oname[ode->d_name.len] = '\0';
	memcpy(rqst->msg.nname, nde->d_name.name, nde->d_name.len);
	rqst->msg.nname[nde->d_name.len] = '\0';

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_RENAME);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (rply) kfree(rply);

//	iput(odir);	// gets "semaphore still in use" message; still aliases
//	iput(ndir);	// gets "semaphore still in use" message; still aliases
//	dput(ode);	// produces negative d_count during mv operation
//	dput(nde);
	return result;
}

static int nkfs_readlink(struct dentry *de, char *buffer, int len)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_readlink_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_readlink_rply_t	msg;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_readlink_rqst_t);
	int	result;
	struct inode	*ip = de->d_inode;

#ifdef DEBUG
	printk("NKFS: read_link: entry \n"); 
#endif

	if (ip == NULL) return -ENOENT;
	if (!S_ISLNK(ip->i_mode) || (ip->u.generic_ip == NULL)) {
//		dput(de);
		return -EINVAL;
	}

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		printk("nkfs_readlink: No connection.\n");
		dput(de);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	rqst->msg.nodeid = nkfsi->nodeid;

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_READLINK);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (result == 0)  {
		copy_to_user(buffer, rply->msg.data, rply->msg.count);
		result = rply->msg.count;
	}

	if (rply) kfree(rply);

	if (ip->i_count < 1) {
		printk("nkfs_read_link: Reputting IP.\n");
	}
//	iput(ip);
	return result;
}

static struct dentry * nkfs_follow_link(struct dentry *dip, 
					struct dentry *drip, 
					unsigned int mode )
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_readlink_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_readlink_rply_t	msg;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_readlink_rqst_t);

	int	err;

	struct inode	*ip = dip->d_inode;
	struct inode	*rip = drip->d_inode;

#ifdef DEBUG
	printk("NKFS: follow_link: dip=%08X\n", dip); 
	printk("NKFS: follow_link: ip=%08X\n", ip); 
	printk("NKFS: follow_link: dip's name is %s\n", dip->d_name.name); 
	printk("NKFS: follow_link: d_count of dip=%d\n", dip->d_count); 
	printk("NKFS: follow_link: d_count of drip=%d\n", drip->d_count); 
#endif

	rip = NULL;

	if ((ip == NULL) || (ip->u.generic_ip == NULL)) {
		printk("nkfs_follow_link: NULL ip.\n");
		dput(drip);
		return dget(dip);
	}

	if (!S_ISLNK(ip->i_mode)) {
		if (ip->i_count < 1)
			printk("nkfs_follow_link: Returning IP 2\n");
//		dput(drip);
		return dget(dip);
	}
	
	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		printk("nkfs_follow_link: No connection.\n");
		dput(drip);
		return dget(dip);
	}
	
	rqst = (struct request *) make_nkfs_request(reqlen);
	rqst->msg.nodeid = nkfsi->nodeid;
	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_FOLLOWLINK);

	if (rply) {
		err = rply->hdr.result;
	} else { 
		err = -EINTR;
	}

	if (ip->i_count < 1) {
		printk("nkfs_follow_link: Reputting IP.\n");
	}

	if (err == 0) 
	{
		current->link_count++;
		if (ip->i_count < 1) {
			printk("nkfs_follow_link: Reputting DIRP.\n");
		}

		drip = lookup_dentry(rply->msg.data, drip, mode);
		current->link_count--;
	} else {
//		dput(drip);
	}

	if (rply) kfree(rply);

	if (err)
		drip = ERR_PTR(err);

	return drip;
	
}

static ssize_t
nkfs_read(struct file *fp, char *buffer, size_t count, loff_t *lf)
{
struct request {
	nkfs_rqst_hdr_t		hdr;
	nkfs_read_rqst_t	msg;
}	*rqst;
struct reply {
	nkfs_rply_hdr_t		hdr;
	nkfs_read_rply_t	msg;
}	*rply;
nkfs_inode_t		*nkfsi;
nkfs_super_block_t	*nkfssb;
int	reqlen = sizeof(nkfs_read_rqst_t);
int	total = 0;
int	result;
int	reqcnt;
int	repcnt;
int	pos;
struct dentry *de = fp->f_dentry;
struct inode *ip = de->d_inode;

#ifdef DEBUG
	printk("NKFS:read: entry");
#endif


	if (ip == NULL) return -ENOENT;
	if (ip->u.generic_ip == NULL)
		return -EINVAL;

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		return -EINVAL;
	}

	pos = fp->f_pos;
	while (count > 0) {
		reqcnt = (count > NKFS_MAX_READ) ? NKFS_MAX_READ : count;
		rqst = (struct request *) make_nkfs_request(reqlen);
		rqst->msg.nodeid = nkfsi->nodeid;
		rqst->msg.offset = pos;
		rqst->msg.count = reqcnt;
		rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_READ);

		if (rply == NULL) return -EINTR;
		
		if ((result = rply->hdr.result) != 0) {
			kfree(rply);
			return result;
		}

		repcnt = rply->msg.count;
		
		copy_to_user(buffer, rply->msg.data, repcnt);

		pos += repcnt;
		buffer += repcnt;
		total += repcnt;
		count -= repcnt;
		
		kfree(rply);
		if (repcnt < reqcnt) break;
	}

	fp->f_pos = pos;
	return total;
}

static int 
nkfs_write(struct file *fp, const char *buffer, size_t count, loff_t *lp)
{
struct request {
	nkfs_rqst_hdr_t		hdr;
	nkfs_write_rqst_t	msg;
}	*rqst;

struct reply {
	nkfs_rply_hdr_t		hdr;
	nkfs_write_rply_t	msg;
}	*rply;

nkfs_inode_t		*nkfsi;
nkfs_super_block_t	*nkfssb;
int	reqlen = sizeof(nkfs_write_rqst_t);
int	reqcnt;
int	repcnt;
int	result;
int	pos;
int	total = 0;
struct dentry *de = fp->f_dentry;
struct inode *ip = de->d_inode;

#ifdef DEBUG
	printk("NKFS:write: entry\n");
#endif


	if (ip == NULL) return -ENOENT;
	if (!S_ISREG(ip->i_mode) || (ip->u.generic_ip == NULL))
		return -EINVAL;

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		return -EINVAL;
	}

	pos = fp->f_pos;
	if ((fp->f_flags & O_APPEND) && (ip->i_size > pos))
		pos = ip->i_size;
	
	while (count > 0) {
		reqcnt = (count > NKFS_MAX_WRITE) ? NKFS_MAX_WRITE : count;
		rqst = (struct request *) make_nkfs_request(reqlen+reqcnt);
		rqst->msg.nodeid = nkfsi->nodeid;
		rqst->msg.offset = pos;
		rqst->msg.count = reqcnt;
		rqst->msg.data[0] = '1';
		copy_from_user(rqst->msg.data, buffer, reqcnt);

		rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen+reqcnt, NKFS_WRITE);

		if (rply == NULL) return -EINTR;
		
		if ((result = rply->hdr.result) != 0) {
			kfree(rply);
			return result;
		}

		repcnt = rply->msg.count;
		total += repcnt;
		buffer += repcnt;
		pos += repcnt;
		count -= repcnt;
		
		nkfs_refresh_inode(ip, &(rply->msg.nodeinfo));
		kfree(rply);
		if (repcnt < reqcnt) break;
	}

	fp->f_pos = pos;
	if (pos > ip->i_size) ip->i_size = pos;
	return total;
}

static unsigned long
nkfs_mmap_no_page(struct vm_area_struct *vma, unsigned long address, 
		int no_share)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_read_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_read_rply_t	msg;
	}	*rply;
	
	struct file 	*file = vma->vm_file;
	struct dentry 	*dentry = file->f_dentry;
	struct inode 	*ip = dentry->d_inode;

	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	char	*buffer;
	char	*tmp;
	int	reqlen = sizeof(nkfs_read_rqst_t);
	unsigned int	clear;
	int	pos;
	int	count;
	int	reqcnt;
	int	repcnt;
	unsigned long	page;

#ifdef DEBUG
	printk("NKFS: mmap_nopage \n");
#endif

	if (ip == NULL) return -ENOENT;
	if (!S_ISREG(ip->i_mode) || (ip->u.generic_ip == NULL))
		return -EINVAL;

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		return -EINVAL;
	}

	page = __get_free_page(GFP_KERNEL);
	if (page == 0L) return 0L;

	buffer = (char *)page;
	address &= PAGE_MASK;
	pos = (address - vma->vm_start) + vma->vm_offset;

	clear = 0;
	if ((address + PAGE_SIZE) > vma->vm_end)
		clear = (address + PAGE_SIZE) - vma->vm_end;

	count = PAGE_SIZE - clear;
	while (count > 0) {
		reqcnt = (count > NKFS_MAX_READ) ? NKFS_MAX_READ : count;
		rqst = (struct request *) make_nkfs_request(reqlen);
		rqst->msg.nodeid = nkfsi->nodeid;
		rqst->msg.offset = pos;
		rqst->msg.count = reqcnt;
		rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_READ);

		if (rply == NULL) return -EINTR;
		
		if (rply->hdr.result) {
			kfree(rply);
			break;
		}

		repcnt = rply->msg.count;
		memcpy(buffer, rply->msg.data, repcnt);
		pos += repcnt;
		buffer += repcnt;
		count -= repcnt;
		
		kfree(rply);
		if (repcnt < reqcnt) break;
	}
	tmp = (char *)((page + PAGE_SIZE) - clear);
	memset(tmp, '\0', clear);

	return page;
}

	
static struct vm_operations_struct nkfs_vm_ops = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	nkfs_mmap_no_page,
	NULL,
	NULL,
	NULL,
};


static int nkfs_mmap(struct file *fp, struct vm_area_struct *vp)
{
	struct inode *ip = fp->f_dentry->d_inode;	

#ifdef DEBUG
	printk("NKFS: mmap \n");
#endif

	if ( (!ip->i_sb) || !S_ISREG(ip->i_mode))
		return -EACCES;
	if (vp->vm_flags & VM_SHARED) 
		return -EINVAL;
	if (!IS_RDONLY(ip))
		ip->i_atime = CURRENT_TIME;

	vp->vm_ops = &nkfs_vm_ops;
	return 0;

}

static int nkfs_readdir(struct file *fp, void *dp, filldir_t fillp)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_readdir_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_readdir_rply_t	msg;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	nkfs_readdir_state_t	*state;
	nkfs_dirent_t		*nextdir;
	long			pos;
	int			result;
	int			reqlen;
	struct inode *ip = fp->f_dentry->d_inode;

#ifdef DEBUG
	printk("NKFS: readdir \n");
#endif

	if ((ip == NULL) || !S_ISDIR(ip->i_mode) || (fp == NULL)) {
		return -EBADF;
	}

	for (;;) {
		pos = fp->f_pos;
		switch (pos) {
		case 0:
			if (fillp(dp, ".", strlen("."), pos, ip->i_ino) < 0)
				return 0;
			fp->f_pos++;
			continue;
			break;

		case 1:
			if (fillp(dp, "..", strlen(".."), pos, 23) < 0)
				return 0;
			fp->f_pos++;
			continue;
			break;

		case 2:
			state = (nkfs_readdir_state_t *) 
				kmalloc(sizeof(*state), GFP_ATOMIC);
			state->signature = NKFS_READDIR_SIG;
			state->reply = NULL;
			state->next = NULL;
			state->buffer = NULL;
			fp->private_data = state;
			fp->f_pos++;
			continue;
			break;
	
		default:
			break;
		}

		state = (nkfs_readdir_state_t *)(fp->private_data);
		if ((state == NULL) || (state->signature != NKFS_READDIR_SIG)) 
			return -EBADF;
	
		if ((state->reply != NULL) && (state->reply->count > 0)
				&& (state->next != NULL)) {
			nextdir = (nkfs_dirent_t *)(state->next);
			if (fillp(dp, nextdir->name, strlen(nextdir->name), 
					pos, nextdir->ino) < 0)
				return 0;
			fp->f_pos++;
			state->next += nextdir->entrylen;
			state->reply->count -= nextdir->entrylen;
			continue;
		}

		nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
		if (nkfsi == NULL) return -EINVAL;
		if (((nkfssb = nkfsi->nkfssb) == NULL) || 
			(nkfssb->driver == NULL)) {
			return -EIO;
		}

		reqlen = sizeof(nkfs_readdir_rqst_t);
		rqst = (struct request *) make_nkfs_request(reqlen);
		rqst->msg.nodeid = nkfsi->nodeid;
		if (pos == 3) {
			/* It is the first time */
			rqst->msg.handle = 0L;
			rqst->msg.verb = NKFS_READDIR_START;
		} else if (state->buffer) {
			/* It is a follow on */
			rqst->msg.handle = state->reply->handle;
			rqst->msg.verb = NKFS_READDIR_MORE;
			kfree(state->buffer);
		} else {
			/* We are done */
			return 0;
		}
	
		state->reply = NULL;
		state->buffer = NULL;
		state->next = NULL;
		rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
				(nkfs_rqst_t *)rqst, reqlen, NKFS_READDIR);

		if (rply == NULL) return -EINTR;
		if ((result = rply->hdr.result) != 0) {
			kfree(rply);
			return result;
		}
	
		if (rply->msg.count == 0) {
			kfree(rply);
			return 0;
		}

		state->buffer = rply;
		state->reply = &(rply->msg);
		state->next = rply->msg.buffer;
	}
	return 0;
}

static int 
nkfs_open(struct inode *ip, struct file *fp)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_open_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_open_rqst_t);
	int	result;

#ifdef DEBUG
	printk("NKFS: nkfs_open \n");
#endif

	if ((ip == NULL) || (ip->u.generic_ip == NULL)) {
		return -ENOENT;
	}
	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		return -EIO;
	}

	fp->private_data = NULL;
	rqst = (struct request *) make_nkfs_request(reqlen);
	rqst->msg.nodeid = nkfsi->nodeid;
	rqst->msg.flags = 0;

	if (fp->f_flags & O_RDWR) {
		rqst->msg.flags = (NKFS_READ_ACCESS | NKFS_WRITE_ACCESS);
	}
	if (fp->f_flags & O_RDONLY) {
		rqst->msg.flags |= NKFS_READ_ACCESS;
	}
	if (fp->f_flags & O_WRONLY) {
		rqst->msg.flags |= NKFS_WRITE_ACCESS;
	}
	
	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_OPEN);

	if (rply) {
		result = rply->hdr.result;
	} else {
		result = -EINTR;
	}

	if (rply) kfree(rply);
	return result;
}

static int 
nkfs_release_file(struct inode *ip, struct file *fp)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_release_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_readdir_state_t	*state;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_release_rqst_t);

#ifdef DEBUG
	printk("NKFS: release_file \n");
#endif

	if ((ip == NULL) || (ip->u.generic_ip == NULL))
		return(0);
	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		return(0);
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	rqst->msg.handle = 0L;
	if (fp->private_data != NULL) {
		state = (nkfs_readdir_state_t *)(fp->private_data);
		if (state->signature == NKFS_READDIR_SIG) {
			if (state->reply != NULL)
				rqst->msg.handle = state->reply->handle;
			if (state->buffer != NULL)
				kfree(state->buffer);
			kfree(state);
		}
		fp->private_data = NULL;
	}
	
	rply = (struct reply *) nkfs_rpc(nkfssb->driver, (nkfs_rqst_t *)rqst, 
			reqlen, NKFS_RELEASE);
	if (rply) kfree(rply);
	return(0);
}

static void 
nkfs_truncate(struct inode *ip)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_truncate_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_truncate_rqst_t);
	int	result;
	
#ifdef DEBUG
	printk("NKFS: truncate \n");
#endif

	if ((ip == NULL) || (ip->u.generic_ip == NULL)) 
		return;

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) 
		return;

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	rqst->msg.length = ip->i_size;
	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_TRUNCATE);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (rply) kfree(rply);

	return;
}

static int 
nkfs_permission(struct inode *ip, int flags)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_permission_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_permission_rqst_t);
	int	result;

#ifdef DEBUG
	printk("NKFS: permission \n");
#endif

	if ((ip == NULL) || (ip->u.generic_ip == NULL)) return -ENOENT;

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) 
	{
		printk("NKFS: permission, nkfssb=%08X or nkfssb->driver= is NULL \n",(unsigned int)nkfssb);
		return -EIO;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;
	if (flags & MAY_READ)
		rqst->msg.flags |= NKFS_READ_ACCESS;
	if (flags & MAY_WRITE)
		rqst->msg.flags |= NKFS_WRITE_ACCESS;
	if (flags & MAY_EXEC)
		rqst->msg.flags |= NKFS_EXEC_ACCESS;

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_PERMISSION);

	if (rply)
		result = rply->hdr.result;
	else
		result = -EINTR;

	if (rply) kfree(rply);

	return result;
}



static void
nkfs_driver_closed(void *arg)
{

	nkfs_super_block_t *nkfssb = (nkfs_super_block_t *)arg;
	nkfssb->driver = NULL;

#ifdef DEBUG
	printk("NKFS: Driver closed \n");
#endif
}

static struct super_block *
nkfs_read_super(struct super_block *sb, void *args, int silent)
{
	nkfs_mount_args_t	*margs = (nkfs_mount_args_t *)args;
	nkfs_super_block_t	*nkfssb;
	nkfs_driver_t		*driver;
	struct file		*fp;
	struct inode		*root_inode;
	unsigned long		flags;
	struct file_struct	*zfiles;
	
#ifdef DEBUG
	printk("NKFS: read_super, sb=%08X, args=%08X, silent=%d\n",
			sb, args, silent);
#endif

	MOD_INC_USE_COUNT;

	if ((margs == NULL) || (margs->fd >= NR_OPEN)) {
		sb->s_dev = 0;
		printk("NKFS: read_super, bad margs \n");
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	fp = Gfp;
	driver = (nkfs_driver_t *)(fp->private_data);
	if ((driver == NULL) || (driver->signature != NKFS_DRIVER_SIG)) {
		printk("NKFS: read_super, driver = NULL or bad driver SIG \n");
		sb->s_dev = 0;
		MOD_DEC_USE_COUNT;
		return NULL;
	}
		
	nkfssb = (nkfs_super_block_t *)
			kmalloc(sizeof(nkfs_super_block_t), GFP_ATOMIC);
	if (nkfssb == NULL) {
		printk("NKFS: read_super, kmalloc failure for nkfsuper \n");
		sb->s_dev = 0;
		MOD_DEC_USE_COUNT;
		return NULL;
	}

	nkfssb->sb = sb;
	nkfssb->driver = driver;
	nkfssb->fp = fp;

	lock_super(sb);
	sb->s_op = &nkfs_super_ops;
	sb->s_magic = NKFS_SUPER_MAGIC;
	sb->s_blocksize = margs->blocksize;
	sb->s_blocksize_bits = margs->blocksize_bits;
	sb->u.generic_sbp = nkfssb;

	root_inode = nkfs_make_inode(sb, &(margs->rootnode));

	sb->s_root = d_alloc_root(root_inode, NULL);
	if (sb->s_root == NULL) {
		printk("NKFS: read_super, root_inode = NULL! free nkfssb\n");
		sb->s_dev = 0;
		kfree(nkfssb);
		MOD_DEC_USE_COUNT;
		unlock_super(sb);
		return NULL;
	}
	
	nkfs_register_notifier(driver, nkfs_driver_closed, nkfssb);

	unlock_super(sb);   // does this really need to be there???
	
	return sb;
}


/*
 * This is called from iget() with the read semaphore held. 
 * The global ncpfs_file_info structure has been set up by ncp_iget.
 */
static void nkfs_read_inode(struct inode *inode)
{
	if (S_ISDIR(inode->i_mode) || S_ISREG(inode->i_mode)) 
	{
	}
	else
	{
#ifdef DEBUG
	printk("NKFS: read_inode, inode=%08X mode = %d is \n",inode,inode->i_mode);
#endif
	}
	inode->i_op = &nkfs_inode_ops;
}


static void
nkfs_delete_inode(struct inode *inode)
{

#ifdef DEBUG
	printk("NKFS: delete_inode \n");
	printk("NKFS delete_inode: count=%d\n", inode->i_count);
#endif

	if (S_ISDIR(inode->i_mode)) {
	}

	clear_inode(inode);
}


extern void
nkfs_send(nkfs_driver_t *, nkfs_rqst_t *, unsigned long, unsigned long);

static int 
nkfs_notify_change(struct dentry *de, struct iattr *iap)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_notify_chg_rqst_t	msg;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_notify_chg_rply_t	msg;
	}	*rply;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_notify_chg_rqst_t);
	int	result;
	struct inode *ip = de->d_inode;  // new for 2.2

#ifdef DEBUG
	printk("NKFS: notify_change \n");
#endif

	if ((ip == NULL) || (ip->u.generic_ip == NULL)) 
		return 0;
	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);
	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL))
	{
		printk("NKFS: notify_change, nkfssb=%08X or driver=NULL \n",(unsigned int)nkfssb);
		return 0;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	
	rqst->msg.nodeid = nkfsi->nodeid;

	if (iap->ia_valid & ATTR_MODE)
		rqst->msg.mode = iap->ia_mode;
	else
		rqst->msg.mode = -1;
		
	if (iap->ia_valid & ATTR_UID)
		rqst->msg.uid = iap->ia_uid;
	else
		rqst->msg.uid = -1;

	if (iap->ia_valid & ATTR_GID)
		rqst->msg.gid = iap->ia_gid;
	else
		rqst->msg.gid = -1;

	if (iap->ia_valid & ATTR_MTIME)
		rqst->msg.mtime = iap->ia_mtime;
	else
		rqst->msg.mtime = -1;

	if (iap->ia_valid & ATTR_ATIME)
		rqst->msg.atime = iap->ia_atime;
	else
		rqst->msg.atime = -1;

	if ((iap->ia_valid & ATTR_SIZE) && (S_ISREG(ip->i_mode)))
		rqst->msg.size = iap->ia_size;
	else
		rqst->msg.size = -1;

	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_NOTIFY_CHANGE);

	if (rply) 
		result = rply->hdr.result;
	else  
		result = -EINTR;

	if (result == 0) nkfs_refresh_inode(ip, &(rply->msg.nodeinfo));
	if ((iap->ia_valid & ATTR_SIZE) && (S_ISREG(ip->i_mode)))
		ip->i_size = iap->ia_size;

	if (rply) kfree(rply);

	return result;
}

static void 
nkfs_put_inode(struct inode *ip)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
		nkfs_put_inode_rqst_t	msg;
	}	*rqst;
	nkfs_inode_t		*nkfsi;
	nkfs_super_block_t	*nkfssb;
	int	reqlen = sizeof(nkfs_put_inode_rqst_t);

#ifdef DEBUG
	printk("NKFS: put_inode \n");
#endif

	if (ip == NULL) 
	{
		printk("NKFS: put_inode, IP = NULL \n");
		return;
	}

	if (ip->u.generic_ip == NULL) {
		printk("NKFS: put_inode, generic_ip is NULL.\n");
		return;
	}

	nkfsi = (nkfs_inode_t *)(ip->u.generic_ip);

	if (((nkfssb = nkfsi->nkfssb) == NULL) || (nkfssb->driver == NULL)) {
		printk("NKFS: put_inode, nkfssb = %08X or driver = NULL.\n",(unsigned int)nkfssb);
		kfree(nkfsi);
		return;
	}

	rqst = (struct request *) make_nkfs_request(reqlen);
	rqst->msg.nodeid = nkfsi->nodeid;

	nkfs_send(nkfssb->driver, (nkfs_rqst_t *)rqst, reqlen, NKFS_PUT_INODE);

//	clear_inode(ip);

	if (ip->i_count == 1) {
		ip->i_nlink = 0;
	}

//	if (nkfsi)
//		kfree(nkfsi);

	return;
}

static void 
nkfs_put_super(struct super_block *sb)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
	}	*rply;
	nkfs_super_block_t	*nkfssb;
	nkfs_driver_t		*driver;
	int	reqlen = 0;

#ifdef DEBUG
	printk("NKFS: put_super \n");
#endif

	if ((sb == NULL) || (sb->u.generic_sbp == NULL)) {
		MOD_DEC_USE_COUNT;
		printk("NKFS: put_super, sb=NULL \n");
		return;
	}
	nkfssb = (nkfs_super_block_t *)(sb->u.generic_sbp);
	driver = nkfssb->driver;

//	lock_super(sb);
	sb->s_dev = 0;
//	unlock_super(sb);

	if (driver != NULL) {
		(void) nkfs_unregister_notifier(driver);

		rqst = (struct request *) make_nkfs_request(reqlen);
		rply = (struct reply *) nkfs_rpc(driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_PUT_SUPER);
		if (rply) kfree(rply);
	}

	kfree(nkfssb);
	MOD_DEC_USE_COUNT;

	return;
}

static int 
nkfs_statfs(struct super_block *sb, struct statfs *buf, int count)
{
	struct request {
		nkfs_rqst_hdr_t		hdr;
	}	*rqst;
	struct reply {
		nkfs_rply_hdr_t		hdr;
		nkfs_statfs_rply_t	msg;
	}	*rply;
	nkfs_super_block_t	*nkfssb;
	nkfs_statfs_rply_t	statstr;
	int	reqlen = 0;
	struct statfs	attr;	


#ifdef DEBUG
	printk("NKFS: statfs \n");
#endif

	if ((sb == NULL) || (sb->u.generic_sbp == NULL))
		return(0); 

	nkfssb = (nkfs_super_block_t *)(sb->u.generic_sbp);
	if ((nkfssb == NULL) || (nkfssb->driver == NULL))
	{
		return(0);
	}
	rqst = (struct request *) make_nkfs_request(reqlen);
	rply = (struct reply *) nkfs_rpc(nkfssb->driver, 
			(nkfs_rqst_t *)rqst, reqlen, NKFS_STATFS);
	if ((rply == NULL) || (rply->hdr.result < 0)) {
		statstr.bavail =
		statstr.bfree =
		statstr.blocks =
		statstr.namelen = 0L;
		statstr.bsize = 1024;
	} else {
		statstr = rply->msg;
	}

	attr.f_type = NKFS_SUPER_MAGIC;
	attr.f_bavail = statstr.bavail;
	attr.f_bsize = statstr.bsize;
	attr.f_blocks = statstr.blocks;
	attr.f_bfree = statstr.bfree;
	attr.f_namelen = statstr.namelen;
	attr.f_files = 0L;
	attr.f_ffree = 0L;
	
	if (rply) kfree(rply);

	copy_to_user_ret(buf, &attr, count, -EFAULT);
	return(0);
}

static struct file_system_type nkfs_fs_type = {
	"nkfs", 
	0, 
	nkfs_read_super, 
	NULL
};

int 
init_nkfsfs(void)
{

#ifdef DEBUG
	printk("NKFS: init_nkfsfs \n");
#endif

	register_filesystem(&nkfs_fs_type);
	return 0;
}

void
cleanup_nkfsfs(void)
{
#ifdef DEBUG
	printk("NKFS: cleanup_nkfsfs \n");
#endif
	unregister_filesystem(&nkfs_fs_type);
}

